@@ -4,6 +4,9 @@
   (c) Simon Marlow 1990-1993
   (c) Albert Graef 1994
 
+  modified 7-1997 by strauman@sun6hft.ee.tu-berlin.de to add
+  different enhancements (see README-NEW).
+
   Functions & data for creating & maintaining  the application window
 -----------------------------------------------------------------------------*/
 
@@ -25,6 +28,11 @@
 
 #define APPWIDTH 300
 
+#ifdef ENHANCE_TRANSLATIONS
+#undef XtNtranslations
+#define XtNtranslations ""
+#endif
+
 /*-----------------------------------------------------------------------------
   PUBLIC DATA 
 -----------------------------------------------------------------------------*/
@@ -34,6 +42,7 @@ app_popup_widget, *app_popup_items, app_popup_widget1;
 int n_appst = 0;
 char **appst = NULL;
 
+
 /*-----------------------------------------------------------------------------
   STATIC DATA 
 -----------------------------------------------------------------------------*/
@@ -51,6 +60,8 @@ static MenuItemRec app_popup_menu[] = {
   { "select all", "Select all", appSelectAllCb },
   { "deselect all", "Deselect all", appDeselectCb },
   { "line4", NULL, NULL },
+  { "about", "About xfm...", aboutCb },
+  { "line5", NULL, NULL },
   { "quit", "Quit", appCloseCb },
 };
 
@@ -105,7 +116,8 @@ static Arg icon_toggle_args[] = {
   { XtNbitmap, (XtArgVal) NULL },
   { XtNtranslations, (XtArgVal) NULL },
   { XtNwidth, 0 },
-  { XtNheight, 0 }
+  { XtNheight, 0 },
+  { XtNforeground, (XtArgVal) 0 },
 };
 
 static Arg icon_label_args[] = {
@@ -122,6 +134,8 @@ static Arg icon_label_args[] = {
   Translation tables
 -----------------------------------------------------------------------------*/
 
+#ifndef ENHANCE_TRANSLATIONS
+
 static char app_translations[] = "\
     <Enter>             : appMaybeHighlight()\n\
     <Leave>             : unhighlight()\n\
@@ -136,6 +150,8 @@ static char iconbox_translations[] = "\
     <Btn3Up>            : dummy()\n\
     <Btn3Down>          : appPopup()\n";
 
+#endif
+
 /*-----------------------------------------------------------------------------
   Action tables
 -----------------------------------------------------------------------------*/
@@ -180,6 +196,7 @@ static int longestName()
   return longest;
 }
 
+
 /*-----------------------------------------------------------------------------
   PUBLIC FUNCTIONS
 -----------------------------------------------------------------------------*/
@@ -269,16 +286,20 @@ Pixmap defaultIcon(char *name, char *directory, char *fname)
 
 void createApplicationWindow()
 {
+#ifndef ENHANCE_TRANSLATIONS
   XtTranslations t;
+#endif
 
   /* Add new actions and parse the translation tables */
   XtAppAddActions(app_context, app_actions, XtNumber(app_actions));
 
+#ifndef ENHANCE_TRANSLATIONS
   t = XtParseTranslationTable(app_translations);
   icon_toggle_args[3].value = (XtArgVal) t;
 
   t = XtParseTranslationTable(iconbox_translations);
   icon_box_args[1].value = (XtArgVal) t;
+#endif
 
   icon_label_args[3].value = (XtArgVal) resources.icon_font;
 
@@ -299,7 +320,11 @@ void createApplicationWindow()
 		       GrabModeAsync, GrabModeAsync);
 
   /* create the form */
+#ifndef ENHANCE_TRANSLATIONS
   aw.form = XtCreateManagedWidget("form", formWidgetClass, aw.shell,
+#else
+  aw.form = XtCreateManagedWidget("awform", formWidgetClass, aw.shell,
+#endif
 				  form_args, XtNumber(form_args) );
   
   /* create the viewport */
@@ -346,6 +371,9 @@ void createApplicationDisplay()
     icon_label_args[4].value = (XtArgVal) width;
     icon_toggle_args[4].value = (XtArgVal) width;
     icon_toggle_args[5].value = (XtArgVal) resources.app_icon_height;
+#ifdef ENHANCE_SELECTION
+    icon_toggle_args[6].value = (XtArgVal) resources.highlight_pixel;
+#endif
     
     for (i=0; i < aw.n_apps; i++) {
       Pixel back;
@@ -383,6 +411,20 @@ void updateApplicationDisplay()
 {
   XtDestroyWidget(aw.icon_box);
   createApplicationDisplay();
+  setApplicationWindowName();
+}
+
+/*---------------------------------------------------------------------------*/
+
+void setApplicationWindowName()
+{
+  char app_file_name[MAXPATHLEN], *p;
+  if ((p = strrchr(resources.app_file, '/')))
+    strcpy(app_file_name, p+1);
+  else
+    strcpy(app_file_name, resources.app_file);
+  XSetIconName(XtDisplay(aw.shell), XtWindow(aw.shell), app_file_name);
+  XStoreName(XtDisplay(aw.shell), XtWindow(aw.shell), app_file_name);
 }
 
 /*---------------------------------------------------------------------------*/
@@ -396,7 +438,7 @@ void readApplicationBitmaps()
     if (!aw.apps[i].icon[0])
       aw.apps[i].icon_bm = defaultIcon(aw.apps[i].name, aw.apps[i].directory,
 				       aw.apps[i].fname);
-    else if ((aw.apps[i].icon_bm = readIcon(aw.apps[i].icon)) == None) {
+    else if ((aw.apps[i].icon_bm = readIcon(aw.apps[i].icon,0,0)) == None) {
       fprintf(stderr, "%s: can't read icon for application %s\n",
 	      progname, aw.apps[i].name);
       aw.apps[i].icon_bm = defaultIcon(aw.apps[i].name, aw.apps[i].directory,
@@ -503,7 +545,7 @@ void installApplication(char *name, char *directory, char *fname, char *icon,
 
   if (!*icon)
     icon_bm = defaultIcon(name, directory, fname);
-  else if ((icon_bm = readIcon(icon)) == None) {
+  else if ((icon_bm = readIcon(icon,0,0)) == None) {
     error("Can't read icon for", name);
     icon_bm = defaultIcon(name, directory, fname);
   } else
@@ -533,7 +575,7 @@ void replaceApplication(AppRec *app, char *name, char *directory, char *fname,
 
   if (!*icon)
     icon_bm = defaultIcon(name, directory, fname);
-  else if ((icon_bm = readIcon(icon)) == None) {
+  else if ((icon_bm = readIcon(icon,0,0)) == None) {
     error("Can't read icon for", name);
     icon_bm = defaultIcon(name, directory, fname);
   } else
