@@ -10,11 +10,15 @@
   modified 1-29-95 by rodgers@lvs-emh.lvs.loral.com (Kevin M. Rodgers)
   to add filtering of icon/text directory displays by a filename filter.
 
+  modified 7-1997 by strauman@sun6hft.ee.tu-berlin.de to add
+  different enhancements (see README-NEW).
+
 -----------------------------------------------------------------------------*/
 
 #include <pwd.h>
 #include <time.h>
 #include <string.h>
+#include <ctype.h>
 
 #ifdef _AIX
 #include <sys/resource.h>
@@ -31,9 +35,33 @@
 #include <X11/Xaw/Toggle.h>
 #include <X11/Xaw/Label.h>
 
+#include "TextFileList.h"
+#include "IconFileList.h"
+
 #include "Am.h"
 #include "Fm.h"
 
+#ifdef ENHANCE_HISTORY
+#include "FmHistory.h"
+#endif
+
+#ifdef ENHANCE_SELECTION
+#include "FmSelection.h"
+#endif
+
+#ifdef ENHANCE_TRANSLATIONS
+/* This is a hack in order not to change the resource tables
+ * because I don't want to re-index all references to these tables
+ * (would have been nicer, if some symbols were defined)
+ */
+#undef  XtNtranslations
+#define XtNtranslations ""
+#endif
+
+#ifdef ENHANCE_LOG
+#include "FmLog.h"
+#endif
+
 #define FW_WIDTH 400
 #define FW_HEIGHT 300
 #define TEXT_PADDING 10
@@ -43,6 +71,10 @@
   PUBLIC DATA                                       
 -----------------------------------------------------------------------------*/
 
+#ifdef ENHANCE_HISTORY
+extern HistoryList path_history;
+#endif
+
 FileWindowList file_windows = NULL;
 FileWindowRec *popup_fw;
 Widget file_popup_widget, *file_popup_items;
@@ -62,12 +94,14 @@ static MenuItemRec file_popup_menu[] = {
   { "edit", "Edit", fileEditCb },
   { "view", "View", fileViewCb },
   { "line1", NULL, NULL },
+  { "rename", "Rename...", renamePopup },
+  { "line2", NULL, NULL },
   { "move", "Move...", movePopup },
   { "copy", "Copy...", copyPopup },
   { "link", "Link...", linkPopup },
-  { "line2", NULL, NULL },
-  { "delete", "Delete", deleteItems },
   { "line3", NULL, NULL },
+  { "delete", "Delete", deleteItems },
+  { "line4", NULL, NULL },
   { "info", "Information...", infoPopup },
   { "chmod", "Permissions...", chmodPopup }
 };
@@ -75,12 +109,14 @@ static MenuItemRec file_popup_menu[] = {
 static MenuItemRec dir_popup_menu[] = {
   { "open", "Open", fileOpenCb },
   { "line1", NULL, NULL },
+  { "rename", "Rename...", renamePopup },
+  { "line2", NULL, NULL },
   { "move", "Move...", movePopup },
   { "copy", "Copy...", copyPopup },
   { "link", "Link...", linkPopup },
-  { "line2", NULL, NULL },
-  { "delete", "Delete", deleteItems },
   { "line3", NULL, NULL },
+  { "delete", "Delete", deleteItems },
+  { "line4", NULL, NULL },
   { "info", "Information...", infoPopup },
   { "chmod", "Permissions...", chmodPopup }
 };
@@ -97,7 +133,14 @@ static MenuItemRec file_menu[] = {
   { "select", "Select...", selectPopup },
   { "select all", "Select all", fileSelectAllCb },
   { "deselect all", "Deselect all", fileDeselectCb },
+#ifdef ENHANCE_SELECTION
+  { "own Selection", "Own Selection", selectionOwnCb },
+#endif
   { "line4", NULL, NULL },
+  { "xterm", "Xterm", xtermCb },
+  { "line5", NULL, NULL },
+  { "about", "About xfm...", aboutCb },
+  { "line6", NULL, NULL },
   { "quit", "Quit", appCloseCb },
 };
 
@@ -110,6 +153,9 @@ static MenuItemRec folder_menu[] = {
   { "line2", NULL, NULL },
   { "empty", "Empty", emptyDir },
   { "line3", NULL, NULL },
+#ifdef ENHANCE_MENU
+  { "clone", "Clone", fileCloneCb },
+#endif
   { "close", "Close", fileCloseCb },
 };
 
@@ -128,6 +174,24 @@ static MenuItemRec view_menu[] = {
   { "mix folders/files", "Mix folders/files",
        fileDirsFirstCb },
   { "show hidden files", "Show hidden files", fileShowHiddenCb },
+#ifdef ENHANCE_LOG
+  { "line4", NULL,  NULL },
+  { "show log", "Show log", logPopup },
+#endif
+};
+
+/* Don't change the order of menu entries
+ * (showTxtOptsCb() and updateTxtOpts() depend on it)
+ */
+static MenuItemRec text_opts_menu[]={
+  { "show Inode", "Show Inode", showTxtOptsCb},
+  { "show Type", "Show Type", showTxtOptsCb},
+  { "show Permission", "Show Permission", showTxtOptsCb},
+  { "show Link Count", "Show Link Count", showTxtOptsCb},
+  { "show User", "Show User", showTxtOptsCb},
+  { "show Group", "Show Group", showTxtOptsCb},
+  { "show Size", "Show Size", showTxtOptsCb},
+  { "show MTime", "Show MTime", showTxtOptsCb},
 };
 
 /*-----------------------------------------------------------------------------
@@ -187,9 +251,24 @@ static Arg status_args[] = {
   { XtNjustify, XtJustifyLeft }
 };
 
-static Arg icon_box_args[] = {
-  { XtNwidth, (XtArgVal) 0 },
-  { XtNtranslations, (XtArgVal) NULL }
+static Arg text_list_args[] = {
+  { XtNwidth,  (XtArgVal) 0 },
+  { XtNfont, (XtArgVal)0 },
+  { XtNfiles, (XtArgVal) 0},
+  { XtNnFiles, (XtArgVal) 0},
+  { XtNtranslations, (XtArgVal) NULL},
+  { XtNhighlightColor, (XtArgVal) 0 },
+};
+
+static Arg icon_list_args[] = {
+  { XtNwidth,  (XtArgVal) 0 },
+  { XtNfont, (XtArgVal)0 },
+  { XtNfiles, (XtArgVal) 0},
+  { XtNnFiles, (XtArgVal) 0},
+  { XtNminIconWidth, (XtArgVal) 0},
+  { XtNminIconHeight, (XtArgVal) 0},
+  { XtNtranslations, (XtArgVal) NULL},
+  { XtNhighlightColor, (XtArgVal) 0 },
 };
 
 static Arg tree_box_args[] = {
@@ -198,11 +277,6 @@ static Arg tree_box_args[] = {
   { XtNheight, (XtArgVal) 1 }
 };
 
-static Arg icon_form_args[] = {
-  { XtNdefaultDistance, (XtArgVal) 0 },
-  { XtNwidth, (XtArgVal) 0 }
-};
-
 static Arg tree_form_args[] = {
   { XtNfromHoriz, (XtArgVal) NULL },
   { XtNfromVert, (XtArgVal) NULL },
@@ -211,7 +285,10 @@ static Arg tree_form_args[] = {
   { XtNtop, XtChainTop },
   { XtNbottom, XtChainTop },
   { XtNleft, XtChainLeft },
-  { XtNright, XtChainLeft }
+  { XtNright, XtChainLeft },
+#ifdef VIEWPORT_HACK
+  { XtNresizable, (XtArgVal) True },
+#endif
 };
 
 static Arg icon_toggle_args[] = {
@@ -220,7 +297,8 @@ static Arg icon_toggle_args[] = {
   { XtNbitmap, (XtArgVal) NULL },
   { XtNtranslations, (XtArgVal) NULL },
   { XtNwidth, (XtArgVal) 0 },
-  { XtNheight, (XtArgVal) 0 }
+  { XtNheight, (XtArgVal) 0 },
+  { XtNforeground, (XtArgVal) 0 },
 };
 
 static Arg icon_label_args[] = {
@@ -234,30 +312,6 @@ static Arg icon_label_args[] = {
   { XtNinternalHeight, (XtArgVal) 0 },
 };
 
-static Arg text_label_args[] = {
-  { XtNfromHoriz, (XtArgVal) NULL },
-  { XtNfromVert, (XtArgVal) NULL },
-  { XtNlabel, (XtArgVal) NULL },
-  { XtNfont, (XtArgVal) NULL },
-  { XtNwidth, (XtArgVal) 0 },
-  { XtNjustify, XtJustifyLeft },
-  { XtNhorizDistance, (XtArgVal) TEXT_PADDING },
-  { XtNinternalWidth, (XtArgVal) 0 },
-  { XtNinternalHeight, (XtArgVal) 0 }
-};
-
-static Arg text_toggle_args[] = {
-  { XtNfromHoriz, (XtArgVal) NULL },
-  { XtNfromVert, (XtArgVal) NULL },
-  { XtNlabel, (XtArgVal) NULL },
-  { XtNfont, (XtArgVal) NULL },
-  { XtNwidth, (XtArgVal) 0 },
-  { XtNtranslations, (XtArgVal) NULL },
-  { XtNjustify, XtJustifyLeft },
-  { XtNinternalWidth, (XtArgVal) 0 },
-  { XtNinternalHeight, (XtArgVal) 0 }
-};
-
 static Arg arrow_args[] = {
   { XtNfromHoriz, (XtArgVal) NULL },
   { XtNfromVert, (XtArgVal) NULL },
@@ -288,8 +342,15 @@ static Arg line_args[] = {
   Translation tables
 -----------------------------------------------------------------------------*/
 
+#ifndef ENHANCE_TRANSLATIONS
+#ifdef ENHANCE_HISTORY
+static char label_translations_s[] = "\
+  <Btn3Down>	      : FmUpdateHistory(fm_history) MenuPopup(fm_history)\n\
+  <Btn1Up>(2)         : fileRefresh()\n";
+#else
 static char label_translations_s[] = "\
   <Btn1Up>(2)         : fileRefresh()\n";
+#endif
 
 static char tree_translations_s[] = "\
   <Enter>             : fileHighlight()\n\
@@ -300,45 +361,34 @@ static char tree_translations_s[] = "\
   <Btn2Down>,<Leave>  : fileBeginDrag(2,copy)\n\
   <Btn3Down>          : dirPopup()\n";
 
-
-static char dir_translations_s[] = "\
-  <Enter>             : fileHighlight()\n\
-  <Leave>             : resetCursor()\n\
+static char file_list_translations_s[] = "\
+  <Enter>	      : dispatchDirExeFile(fileHighlight,fileHighlight,fileMaybeHighlight,\"\")\n\
+  <Leave>	      : dispatchDirExeFile(resetCursor,resetCursor,unhighlight,\"\")\n\
+  <Btn1Up>(2)         : dispatchDirExeFile(fileOpenDir,fileExecFile,fileExecAction,\"\")\n\
   <Btn1Down>,<Btn1Up> : fileSelect()\n\
-  <Btn1Down>,<Leave>  : fileBeginDrag(1,move)\n\
-  <Btn1Up>(2)         : fileOpenDir()\n\
+  <Btn1Down>,<Motion> : fileBeginDrag(1,move)\n\
   <Btn2Down>,<Btn2Up> : fileToggle()\n\
-  <Btn2Down>,<Leave>  : fileBeginDrag(2,copy)\n\
-  <Btn3Down>          : dirPopup()\n";
+  <Btn2Down>,<Motion> : fileBeginDrag(2,copy)\n\
+  <Btn3Down>          : dispatchDirExeFilePopup(dirPopup,filePopup,filePopup,\"\")\n\
+	<Key>Delete   : notify(*file*delete)\n\
+	<Key>BackSpace: notify(*file*delete)\n\
+	<Key>m	      : notify(*file*move)\n\
+	<Key>n	      : notify(*file*new)\n\
+	<Key>x	      : notify(*file*xterm)\n\
+	<Key>q	      : notify(*file*quit)\n\
+	<Key>c	      : notify(*file*copy)\n\
+	<Key>s	      : notify(*file*select)\n\
+	<Key>o	      : notify(\"*file*own selection\")\n\
+	<Key>a	      : notify(\"*file*select all\")\n\
+	<Key>u	      : notify(\"*file*deselect all\")\n\
+	<Key>l	      : notify(*file*link)\n";
 
-static char file_translations_s[] = "\
-  <Enter>             : fileMaybeHighlight()\n\
-  <Leave>             : unhighlight()\n\
-  <Btn1Up>(2)         : fileExecAction()\n\
-  <Btn1Down>,<Btn1Up> : fileSelect()\n\
-  <Btn1Down>,<Leave>  : fileBeginDrag(1,move)\n\
-  <Btn2Down>,<Btn2Up> : fileToggle()\n\
-  <Btn2Down>,<Leave>  : fileBeginDrag(2,copy)\n\
-  <Btn3Down>          : filePopup()\n";
-
-static char exec_translations_s[] = "\
-  <Enter>             : fileHighlight()\n\
-  <Leave>             : resetCursor()\n\
-  <Btn1Up>(2)         : fileExecFile()\n\
-  <Btn1Down>,<Btn1Up> : fileSelect()\n\
-  <Btn1Down>,<Leave>  : fileBeginDrag(1,move)\n\
-  <Btn2Down>,<Btn2Up> : fileToggle()\n\
-  <Btn2Down>,<Leave>  : fileBeginDrag(2,copy)\n\
-  <Btn3Down>          : filePopup()\n";
-
-/* This is a hack to get the icon box to recognise button events */
-static char iconbox_translations_s[] = "\
-    <Btn2Up> : dummy()\n\
-    <Btn3Up> : dummy()\n";
+#endif
   
 static void dummy(Widget w, XEvent *event, String *params, 
 		       Cardinal *num_params) {}
 
+
 static XtActionsRec file_actions[] = {
   { "fileRefresh", fileRefresh },
   { "fileToggle", fileToggle },
@@ -349,14 +399,16 @@ static XtActionsRec file_actions[] = {
   { "fileExecFile", fileExecFile },
   { "fileExecAction", fileExecAction },
   { "resetCursor", resetCursor },
+  { "trackCursor", trackCursor },
   { "fileMaybeHighlight", fileMaybeHighlight },
   { "filePopup", filePopup },
   { "dirPopup", dirPopup },
-  { "dummy", dummy }
+  { "dummy", dummy },
 };
 
-static XtTranslations label_translations, dir_translations, file_translations, 
-  iconbox_translations, tree_translations, exec_translations;
+#ifndef ENHANCE_TRANSLATIONS
+static XtTranslations label_translations, file_list_translations, tree_translations;
+#endif
 
 /*-----------------------------------------------------------------------------
   PRIVATE FUNCTIONS
@@ -374,6 +426,48 @@ static int longestName(FileWindowRec *fw)
   return longest;
 }
 
+static void unmapTxtOptsCb(Widget w, XtPointer cld, XtPointer cad)
+{
+ Widget options_button=(Widget) cld;
+ XtUnmapWidget(options_button);
+}
+
+/* This callback is registered on the callback list of a fileList widget.
+ * It is called by the 'notify()' action.
+ * A pointer to the (null terminated) argument list is passed in the
+ * 'call data' parameter. We use it to determine what to do
+ *
+ * Notify(menu entry path)
+ *
+ * calls the callback registered with an arbitrary menu entry.
+ * the name of the menu entry is resolved by means of
+ * 'XtNameToWidget', starting with the FileWindow's shell.
+ */
+static void fileListCb(Widget w, XtPointer cld, XtPointer cad)
+{
+ FileWindowRec  *fw=(FileWindowRec*)cld;
+ String	        *args=(String*)cad;
+ String	        entry_path,mess;
+ Widget 	entry;
+
+ if ((entry_path=args[0])==0) {
+	XtAppWarning(XtWidgetToApplicationContext(w),
+		"fileListCb (Notify): need 1 arg");
+	return;
+ }
+
+ if ((entry=XtNameToWidget(fw->shell,entry_path))==0) {
+	mess=XtMalloc(256);
+        sprintf(mess,"fileListCb (Notify): menu entry '%s' not found",entry_path);
+	XtAppWarning(XtWidgetToApplicationContext(w),mess);
+	XtFree(mess);
+ 	return;
+ }
+
+ XtCallCallbacks(entry,XtNcallback,0);
+
+}
+
 /*---------------------------------------------------------------------------*/
 
 static int parseType(FILE *fp, char **pattern,
@@ -421,14 +515,19 @@ static int parseType(FILE *fp, char **pattern,
 
 /*---------------------------------------------------------------------------*/
 
-static void readFileBitmaps()
+static void readFileBitmaps(void)
 {
   int i;
 
   for (i=0; i<n_types; i++)
-    if (!types[i].icon[0])
+    if (!types[i].icon[0]) {
       types[i].icon_bm = bm[FILE_BM];
-    else if ((types[i].icon_bm = readIcon(types[i].icon)) == None) {
+      types[i].bm_width=builtin_types[FILE_T].bm_width;
+      types[i].bm_height=builtin_types[FILE_T].bm_height;
+    }
+    else if ((types[i].icon_bm = readIcon(types[i].icon,
+					&types[i].bm_width,
+					&types[i].bm_height)) == None) {
 #ifdef MAGIC_HEADERS
       fprintf(stderr, "%s: can't read icon for type %s%s%s%s%s%s\n", progname,
 	      types[i].magic_type?"<":"",
@@ -440,8 +539,16 @@ static void readFileBitmaps()
       fprintf(stderr, "%s: can't read icon for type %s%s%s\n", progname,
 	      types[i].dir<0?"*":"", types[i].pattern,
 	      types[i].dir>0?"*":"");
+#ifndef ENHANCE_BUGFIX
       types[i].icon_bm = bm[FILE_BM];
 #endif
+#endif
+#ifdef ENHANCE_BUGFIX  
+     /* we need a default icon in any case, whether we use MAGIC_HEADERS or not */
+      types[i].icon_bm = bm[FILE_BM];
+#endif
+      types[i].bm_width=builtin_types[FILE_T].bm_width;
+      types[i].bm_height=builtin_types[FILE_T].bm_height;
     }
 }
 
@@ -520,9 +627,9 @@ static void readFileTypes(String path)
 /*---------------------------------------------------------------------------*/
 
 #ifdef MAGIC_HEADERS
-static TypeRec *fileType(char *name, char *magic_type)
+TypeRec *fileType(char *name, char *magic_type)
 #else
-static TypeRec *fileType(char *name)
+TypeRec *fileType(char *name)
 #endif
 {
   int i, l = strlen(name);
@@ -641,247 +748,85 @@ static int devAction(int d, char *action)
     return 1;
 }
 
-/*---------------------------------------------------------------------------*/
+/*----------------------------------------------------------------------------*/
 
+/* This is new */
 static void createFileIcons(FileWindowRec *fw)
 {
-  int i;
-  Dimension width;
-  FileRec *file;
-  Pixmap icon;
-  char *iconname[4] = {"dir_icon","file_icon","exe_icon","other_icon"};
-  int icontype;
+ Dimension width;
+ int	  nhoriz;
 
-  XtVaGetValues(fw->viewport, XtNwidth, &width, NULL);
-  icon_box_args[0].value = (XtArgVal) width;
 
-  width = longestName(fw);
-  if (width < resources.file_icon_width)
-    width = resources.file_icon_width;
-  icon_form_args[1].value = (XtArgVal) width;
-  icon_toggle_args[4].value = (XtArgVal) width;
-  icon_toggle_args[5].value = (XtArgVal) resources.file_icon_height;
-  icon_label_args[4].value = (XtArgVal) width;
 
-  fw->icon_box = XtCreateWidget("icon box",  boxWidgetClass,
-    fw->viewport, icon_box_args, XtNumber(icon_box_args) );
-
-  for (i=0; i < fw->n_files; i++) {
-    Pixel back;
-    
-    file = fw->files[i];
-    file->icon.form = XtCreateManagedWidget(file->name,
-      formWidgetClass, fw->icon_box, icon_form_args,
-      XtNumber(icon_form_args) );
-
-#ifdef MAGIC_HEADERS
-    /* determine file type first, to allow special items like directories to
-       have custom icons */
-    file->type = fileType(file->name, file->magic_type);
-    if (file->type)
-      icon = (XtArgVal) file->type->icon_bm;
-    else
-      icon = None;
-#else
-    icon = None;
+ XtVaGetValues(fw->viewport, XtNwidth, &width, NULL);
+ icon_list_args[0].value = (XtArgVal) width;
+ icon_list_args[1].value = (XtArgVal) resources.icon_font;
+ icon_list_args[2].value = (XtArgVal) fw->files;
+ icon_list_args[3].value = (XtArgVal) fw->n_files;
+ icon_list_args[4].value = (XtArgVal) resources.file_icon_width;
+ icon_list_args[5].value = (XtArgVal) resources.file_icon_height;
+#ifndef ENHANCE_TRANSLATIONS
+ icon_list_args[6].value = (XtArgVal) file_list_translations;
 #endif
-    icontype = 3;
-
-    /* Symbolic link to non-existent file */
-    if (S_ISLNK(file->stats.st_mode)) {
-      icon_toggle_args[3].value = (XtArgVal) file_translations;
-      if (icon == None)
-        icon = (XtArgVal) bm[BLACKHOLE_BM];
-    }
-    else if (S_ISDIR(file->stats.st_mode)) {
-      icon_toggle_args[3].value = (XtArgVal) dir_translations;
-      if (icon == None)
-	if (file->sym_link)
-	  icon = (XtArgVal) bm[DIRLNK_BM];
-	else if (!strcmp(file->name, ".."))
-	  icon = (XtArgVal) bm[UPDIR_BM];
-        else
-	  icon = (XtArgVal) bm[DIR_BM];
-      icontype = 0;
-    }
-    else if (file->stats.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH)) {
-      icon_toggle_args[3].value = (XtArgVal) exec_translations;
-      if (icon == None)
-	if (file->sym_link)
-	  icon = (XtArgVal) bm[EXECLNK_BM];
-	else
-	  icon = (XtArgVal) bm[EXEC_BM];
-      icontype = 2;
-    }
-    else {
-      icon_toggle_args[3].value = (XtArgVal) file_translations;
-#ifdef MAGIC_HEADERS
-      /* Already got file->type. */
-#else
-      file->type = fileType(file->name);
+#ifdef ENHANCE_SELECTION
+ icon_list_args[7].value = (XtArgVal) resources.highlight_pixel;
 #endif
-      if (icon == None)
-        if (file->type)
-	  icon = (XtArgVal) file->type->icon_bm;
-	else if (file->sym_link)
-	  icon = (XtArgVal) bm[SYMLNK_BM];
-	else
-	  icon = (XtArgVal) bm[FILE_BM];
-      icontype = 1;
-    }
-    icon_toggle_args[2].value = icon;
+ 
+ fw->icon_box = XtCreateWidget("icon list",
+				iconFileListWidgetClass,
+				fw->viewport, icon_list_args,
+				XtNumber(icon_list_args) );
 
-    file->icon.toggle = XtCreateManagedWidget(iconname[icontype],
-      toggleWidgetClass, file->icon.form, icon_toggle_args,
-      XtNumber(icon_toggle_args) );
-    
-    XtVaGetValues(file->icon.toggle, XtNbackground, &back, NULL);
-    XtVaSetValues(file->icon.toggle, XtNborder, (XtArgVal) back, NULL);
+ XtAddCallback(fw->icon_box,XtNcallback,fileListCb,(XtPointer)fw);
 
-    icon_label_args[0].value = (XtArgVal) NULL;
-    icon_label_args[1].value = (XtArgVal) file->icon.toggle;
-    icon_label_args[2].value = (XtArgVal) file->name;
-    file->icon.label = XtCreateManagedWidget("label",
-      labelWidgetClass, file->icon.form, icon_label_args,
-      XtNumber(icon_label_args) );
-  }
+ XtVaGetValues(fw->icon_box,XtNnHoriz,&nhoriz,0);
+ XtVaSetValues(fw->viewport, XtNallowHoriz, nhoriz!=0 , NULL);
 }
 
 /*----------------------------------------------------------------------------*/
 
+/* This is new */
 static void createTextDisplay(FileWindowRec *fw)
 {
-  int i, l;
-  Widget w;
-  Dimension width, m_width, name_w, size_w, perm_w, own_w = 0, date_w;
-  char s[10], name[FILENAME_MAX];
-  struct passwd *pw;
-  char **owners = NULL;
-  FileRec *file;
+ Dimension width;
+ Widget	   options_button;
 
-  XtVaGetValues(fw->viewport, XtNwidth, &width, NULL);
-  icon_box_args[0].value = (XtArgVal) width;
+ XtVaSetValues(fw->viewport, XtNallowHoriz, True, NULL);
 
-  m_width = XTextWidth(resources.icon_font, "m", 1);
-  name_w = longestName(fw) + 2*m_width;
-  size_w = m_width * 7;
-  perm_w = m_width * 9;
-  date_w = m_width * 20;
-  
-  if (resources.show_owner) {
-    owners = (char **) XtMalloc(fw->n_files * sizeof(char *));
-    own_w = 0;
-    for (i=0; i<fw->n_files; i++) { 
-      /* bug fixed by hkarhune@hydra.helsinki.fi - Thanks */
-      if((pw = getpwuid(fw->files[i]->stats.st_uid)) == NULL) {
-	char tmp[11];
- 	
-	sprintf(tmp, "%lu", (unsigned long) fw->files[i]->stats.st_uid);
-	owners[i] = XtNewString(tmp);
-      }
-      else
-	owners[i] = XtNewString(pw->pw_name);
-      l = XTextWidth(resources.icon_font, owners[i], strlen(owners[i]));
-      if (l > own_w)
-	own_w = l;
-    }
-  }
-
-  fw->icon_box = XtCreateWidget("icon box",  boxWidgetClass,
-				fw->viewport, icon_box_args,
-				XtNumber(icon_box_args) );
-
-  for (i=0; i<fw->n_files; i++) {
-    Pixel pix;
-    
-    file = fw->files[i];
-#ifdef MAGIC_HEADERS
-    file->type = fileType(file->name, file->magic_type);
+ XtVaGetValues(fw->viewport, XtNwidth, &width, NULL);
+ text_list_args[0].value = (XtArgVal) width;
+ text_list_args[1].value = (XtArgVal) resources.icon_font;
+ text_list_args[2].value = (XtArgVal) fw->files;
+ text_list_args[3].value = (XtArgVal) fw->n_files;
+#ifndef ENHANCE_TRANSLATIONS
+ text_list_args[4].value = (XtArgVal) file_list_translations;
 #endif
-    if (S_ISDIR(file->stats.st_mode)) {
-      sprintf(name, "[%s]", file->name);
-      text_toggle_args[2].value = (XtArgVal) name;
-      text_toggle_args[5].value = (XtArgVal) dir_translations;
-    }
-    else  {
-      text_toggle_args[2].value = (XtArgVal) file->name;
-      if (file->stats.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH))
-	text_toggle_args[5].value = (XtArgVal) exec_translations;
-      else {
-#ifdef MAGIC_HEADERS
-	/* already got file type */
-#else
-        file->type = fileType(file->name); /* needed for push/drop-actions */
+#ifdef ENHANCE_SELECTION
+ text_list_args[5].value = (XtArgVal) resources.highlight_pixel;
 #endif
-	text_toggle_args[5].value = (XtArgVal) file_translations;
-      }
-    }
 
-    file->icon.form = XtCreateManagedWidget(file->name,
-      formWidgetClass, fw->icon_box, icon_form_args,
-      XtNumber(icon_form_args) );
+ fw->icon_box = XtCreateWidget("text list",
+				textFileListWidgetClass,
+				fw->viewport, text_list_args,
+				XtNumber(text_list_args) );
 
-    text_toggle_args[0].value = (XtArgVal) NULL;
-    text_toggle_args[4].value = (XtArgVal) name_w;
-    w = file->icon.toggle = XtCreateManagedWidget("name", 
-      toggleWidgetClass, file->icon.form, text_toggle_args,
-      XtNumber(text_toggle_args) );
+ XtAddCallback(fw->icon_box,XtNcallback,fileListCb,(XtPointer)fw);
 
-    XtVaGetValues(file->icon.toggle, file->selected?XtNforeground:
-		  XtNbackground, &pix, NULL);
-    XtVaSetValues(file->icon.toggle, XtNborder, (XtArgVal) pix, NULL);
+ options_button=XtParent(XtParent(fw->option_items[0]));
 
-    if (resources.show_length) {
-      sprintf(s, "%ld", (long) file->stats.st_size);
-      text_label_args[0].value = (XtArgVal) w;
-      text_label_args[2].value = (XtArgVal) s;
-      text_label_args[4].value = (XtArgVal) size_w;
-      text_label_args[5].value = (XtArgVal) XtJustifyRight;
-      w = XtCreateManagedWidget("size", labelWidgetClass, 
-				file->icon.form, text_label_args,
-				XtNumber(text_label_args) );
-    }
+ /* Note: we can not pass 'fw' to the procedure on the 'destroyCallback'
+  *       list, because *fw may have been freed already at the time
+  *	  the callback is invoked (see fileCloseCb);
+  */
+ XtAddCallback(fw->icon_box,XtNdestroyCallback,
+		unmapTxtOptsCb,(XtPointer)options_button);
+ 
+ updateTxtOpts(fw);
 
-    if (resources.show_owner) {
-      text_label_args[0].value = (XtArgVal) w;
-      text_label_args[2].value = (XtArgVal) owners[i];
-      text_label_args[4].value = (XtArgVal) own_w;
-      text_label_args[5].value = (XtArgVal) XtJustifyLeft;
-      w = XtCreateManagedWidget("owner", labelWidgetClass, 
-				file->icon.form, text_label_args,
-				XtNumber(text_label_args) );
-    }
-
-    if (resources.show_perms) {
-      makePermissionsString(s, file->stats.st_mode);
-      text_label_args[0].value = (XtArgVal) w;
-      text_label_args[2].value = (XtArgVal) s;
-      text_label_args[4].value = (XtArgVal) perm_w;
-      text_label_args[5].value = (XtArgVal) XtJustifyLeft;
-      w = XtCreateManagedWidget("permissions", labelWidgetClass, 
-				file->icon.form, text_label_args,
-				XtNumber(text_label_args) );
-    }
-
-    if (resources.show_date) {
-      text_label_args[0].value = (XtArgVal) w;
-      text_label_args[2].value = (XtArgVal)ctime(&file->stats.st_mtime);
-      text_label_args[4].value = (XtArgVal) date_w;
-      text_label_args[5].value = (XtArgVal) XtJustifyLeft;
-      w = XtCreateManagedWidget("date", labelWidgetClass, 
-				file->icon.form, text_label_args,
-				XtNumber(text_label_args) );
-    }
-  }
-
-  if (resources.show_owner) {
-    for(i=0; i<fw->n_files; i++)
-      XTFREE(owners[i]);
-    XTFREE(owners);
-  }
+ /* show the text options menu */
+ XtMapWidget(options_button);
 }
 
-
 /*----------------------------------------------------------------------------*/
 
 /* create a directory icon in position specified by horiz & vert */
@@ -903,15 +848,26 @@ static Widget createDirIcon(FileWindowRec *fw, int i, Widget horiz,Widget vert)
   /* create form */
   tree_form_args[0].value = (XtArgVal) horiz;
   tree_form_args[1].value = (XtArgVal) vert;
+#ifdef VIEWPORT_HACK
+  file->icon.form = XtCreateWidget(file->name,
+#else
   file->icon.form = XtCreateManagedWidget(file->name,
+#endif
     formWidgetClass, fw->icon_box, tree_form_args, XtNumber(tree_form_args) );
 
   /* create icon */
   icon_toggle_args[0].value = (XtArgVal) NULL;
   icon_toggle_args[1].value = (XtArgVal) NULL;
   icon_toggle_args[2].value = (XtArgVal) icon;
+#ifdef ENHANCE_SELECTION
+  icon_toggle_args[6].value = (XtArgVal) resources.highlight_pixel;
+#endif
+#ifndef ENHANCE_TRANSLATIONS
   icon_toggle_args[3].value = (XtArgVal) tree_translations;
   file->icon.toggle = XtCreateManagedWidget("icon",
+#else
+  file->icon.toggle = XtCreateManagedWidget("tree_icon",
+#endif
     toggleWidgetClass, file->icon.form, icon_toggle_args,
     XtNumber(icon_toggle_args) );
 
@@ -945,6 +901,11 @@ static void createTreeDisplay(FileWindowRec *fw)
   Pixmap line_bm;
   Dimension width;
   FileList files = fw->files;
+#ifdef VIEWPORT_HACK
+  Boolean force_bars;
+  WidgetList children;
+  Cardinal   num_children;
+#endif
 
   /* find width of icons */
   width = longestName(fw);
@@ -958,11 +919,28 @@ static void createTreeDisplay(FileWindowRec *fw)
   icon_label_args[4].value = (XtArgVal) width;
 
   /* create icon box in viewport */
+#ifdef VIEWPORT_HACK
+  XtVaGetValues(fw->viewport, XtNwidth, &width, XtNforceBars, &force_bars, NULL);
+  /* force creation of scrollbars */
+  XtVaSetValues(fw->viewport, XtNallowHoriz, True, XtNforceBars, True, NULL);
+#else		
+  XtVaSetValues(fw->viewport, XtNallowHoriz, True, NULL);
   XtVaGetValues(fw->viewport, XtNwidth, &width, NULL);
+#endif
   tree_box_args[0].value = (XtArgVal) width;
+#ifdef VIEWPORT_HACK
+  /* widget must be managed by viewport 
+   */
+  fw->icon_box = XtCreateManagedWidget("icon box", formWidgetClass,
+#else
   fw->icon_box = XtCreateWidget("icon box", formWidgetClass,
+#endif
     fw->viewport, tree_box_args, XtNumber(tree_box_args) );
 
+#ifdef VIEWPORT_HACK
+  XawFormDoLayout(fw->icon_box,False);
+#endif
+
   /* The '..' directory is not displayed, and no arrow for '.'  */
   files[1]->icon.form = files[1]->icon.toggle = 
     files[1]->icon.label = NULL;
@@ -1030,6 +1008,13 @@ static void createTreeDisplay(FileWindowRec *fw)
 			      arrow_args, XtNumber(arrow_args) );
     XtAddCallback(vert, XtNcallback, (XtCallbackProc) mainArrowCb, fw);
   }
+#ifdef VIEWPORT_HACK
+  XawFormDoLayout(fw->icon_box,False);
+  XtVaGetValues(fw->icon_box,XtNchildren,&children,XtNnumChildren,&num_children,0);
+  XtManageChildren(children,num_children);
+  if (!force_bars)
+    XtVaSetValues(fw->viewport,XtNforceBars,False,0);
+#endif
 }
 
 /*-----------------------------------------------------------------------------
@@ -1076,21 +1061,17 @@ void umountDev(int d)
 void initFileWindows()
 {
   XtAppAddActions(app_context, file_actions, XtNumber(file_actions));
+#ifndef ENHANCE_TRANSLATIONS
   label_translations = XtParseTranslationTable(label_translations_s);
-  dir_translations = XtParseTranslationTable(dir_translations_s);
-  file_translations = XtParseTranslationTable(file_translations_s);
-  iconbox_translations = XtParseTranslationTable(iconbox_translations_s);
   tree_translations = XtParseTranslationTable(tree_translations_s);
-  exec_translations = XtParseTranslationTable(exec_translations_s);
+  file_list_translations = XtParseTranslationTable(file_list_translations_s);
 
-  icon_box_args[1].value = (XtArgVal) iconbox_translations;
   label_args[9].value = (XtArgVal) label_translations;
+#endif
 
   label_args[3].value = (XtArgVal) resources.label_font;
   status_args[3].value = (XtArgVal) resources.status_font;
   icon_label_args[3].value = (XtArgVal) resources.icon_font;
-  text_toggle_args[3].value = (XtArgVal) resources.icon_font;
-  text_label_args[3].value = (XtArgVal) resources.icon_font;
   shell_args[1].value = (XtArgVal) bm[ICON_BM];
   shell_args[2].value = (XtArgVal) bm[ICONMSK_BM];
 
@@ -1119,7 +1100,7 @@ static FileWindowRec *createFileWindow(String path, String title,
 {
   FileWindowRec *fw;
   char *shell_name;
-  
+
 #ifdef DEBUG_MALLOC
   fprintf(stderr, "entering createFileWindow: %lu\n", malloc_inuse(NULL));
 #endif
@@ -1157,6 +1138,15 @@ static FileWindowRec *createFileWindow(String path, String title,
   fw->do_filter = False;
   fw->dirFilter[0] = '\0';
 
+  fw->showInode=resources.show_inode;
+  fw->showType=resources.show_type;
+  fw->showLinks=resources.show_links;
+  fw->showPermissions=resources.show_perms;
+  fw->showOwner=resources.show_owner;
+  fw->showGroup=resources.show_group;
+  fw->showLength=resources.show_length;
+  fw->showDate=resources.show_date;
+
   shell_name = "file window";
   shell_args[0].value = (XtArgVal) title;
   fw->shell = XtCreatePopupShell(shell_name, topLevelShellWidgetClass,
@@ -1181,11 +1171,18 @@ static FileWindowRec *createFileWindow(String path, String title,
 				(XtPointer) fw);
   fw->view_items = createMenu("view", "View", view_menu, XtNumber(view_menu),
 			      16, fw->button_box, (XtPointer) fw);
+  fw->option_items=createMenu("options","Options",text_opts_menu,XtNumber(text_opts_menu),
+			      16, fw->button_box, (XtPointer) fw);
+  XtSetMappedWhenManaged(XtParent(XtParent(fw->option_items[0])),False);
 
   /* create folder label */
   label_args[0].value = (XtArgVal) fw->button_box;
   label_args[1].value = (XtArgVal) fw->directory;
+#ifndef ENHANCE_TRANSLATIONS
   fw->label = XtCreateManagedWidget("label", labelWidgetClass, fw->form,
+#else
+  fw->label = XtCreateManagedWidget("folderlabel", labelWidgetClass, fw->form,
+#endif
 				    label_args, XtNumber(label_args) );
   
   /* create viewport */
@@ -1222,6 +1219,7 @@ void newFileWindow(String path, DisplayType d, Boolean by_cursor)
   createFileDisplay(fw);
   XtRealizeWidget(fw->shell);
   XSetIconName(XtDisplay(fw->shell), XtWindow(fw->shell), fw->directory);
+  XStoreName(XtDisplay(fw->shell), XtWindow(fw->shell), fw->directory);
   XSetWMProtocols(XtDisplay(fw->shell), XtWindow(fw->shell),
 	  	  &wm_delete_window, 1);
   XtAddEventHandler(fw->shell, (EventMask)0L, True,
@@ -1286,9 +1284,14 @@ void createFileDisplay(FileWindowRec *fw)
     createTextDisplay(fw);
     break;
   }
-
   updateStatus(fw);
 
+#ifdef ENHANCE_HISTORY
+  FmInsertHistoryPath(path_history,fw->directory);
+  if (resources.history_max_n>0)
+    FmChopHistoryList(path_history,resources.history_max_n);
+#endif
+
   XtManageChild(fw->icon_box);
 
 #ifdef DEBUG_MALLOC
@@ -1299,9 +1302,16 @@ void createFileDisplay(FileWindowRec *fw)
 /*---------------------------------------------------------------------------*/
 
 /* Update the display in the viewport */
+#ifdef ENHANCE_SCROLL
+void updateFileDisplay(FileWindowRec *fw, Boolean keep_position)
+#else
 void updateFileDisplay(FileWindowRec *fw)
+#endif
 {
   int d;
+#ifdef ENHANCE_SCROLL
+  Position x,y;
+#endif
 
 #ifdef DEBUG_MALLOC
   fprintf(stderr, "entering updateFileDisplay: %lu\n", malloc_inuse(NULL));
@@ -1311,15 +1321,31 @@ void updateFileDisplay(FileWindowRec *fw)
 
   d = fw->dev;
 
+#ifdef ENHANCE_SCROLL
+  x=y=0;
+  if (fw->icon_box) {
+    if (keep_position)
+      XtVaGetValues(fw->icon_box,XtNx,&x,XtNy,&y,0);
+    XtUnrealizeWidget(fw->icon_box);
+    XtDestroyWidget(fw->icon_box);
+  }
+#else
   if (fw->icon_box)
     XtDestroyWidget(fw->icon_box);
+#endif
 
   freeFileList(fw);
   createFileDisplay(fw);
 
+#ifdef ENHANCE_SCROLL
+  if (keep_position)
+    XawViewportSetCoordinates(fw->viewport,-x,-y);
+#endif
+
   if (d != -1) umountDev(d);
 
   XSetIconName(XtDisplay(fw->shell), XtWindow(fw->shell), fw->directory);
+  XStoreName(XtDisplay(fw->shell), XtWindow(fw->shell), fw->directory);
 
   wakeUp();
 
@@ -1333,6 +1359,9 @@ void updateFileDisplay(FileWindowRec *fw)
 /* resort the icons in the display */
 void reSortFileDisplay(FileWindowRec *fw)
 {
+#ifdef ENHANCE_BUGFIX
+  int i;
+#endif
 #ifdef DEBUG_MALLOC
   fprintf(stderr, "entering resortFileDisplay: %lu\n", malloc_inuse(NULL));
 #endif
@@ -1342,22 +1371,36 @@ void reSortFileDisplay(FileWindowRec *fw)
 
   zzz();
 
-  XtDestroyWidget(fw->icon_box);
-  fw->n_selections = 0;
-  fw->n_bytes_selected = 0;
 
   switch (fw->display_type) {
   case Tree:
+#ifdef ENHANCE_SELECTION
+    FmDisownSelection(fw);
+#endif
+    XtDestroyWidget(fw->icon_box);
+#ifdef ENHANCE_BUGFIX
+    /* when we set n_selections=0 we have to
+     * deselect all the files too! (createTreeDisplay
+     * does not check for the status of file->selected)
+     */
+    for (i=0; i<fw->n_files; i++) {
+	fw->files[i]->selected=False;
+    }
+#endif
+    fw->n_selections = 0;
+    fw->n_bytes_selected = 0;
     sortDirectory(fw->files+2, fw->n_files-2, fw->sort_type, False);
     createTreeDisplay(fw);
     break;
   case Icons:
     sortDirectory(fw->files, fw->n_files, fw->sort_type, fw->dirs_first);
-    createFileIcons(fw);
+    /* force a redisplay (file list, widget and selections are preserved) */
+    XClearArea(XtDisplay(fw->icon_box),XtWindow(fw->icon_box),0,0,0,0,True);
     break;
   case Text:
     sortDirectory(fw->files, fw->n_files, fw->sort_type, fw->dirs_first);
-    createTextDisplay(fw);
+    /* force a redisplay */
+    XClearArea(XtDisplay(fw->icon_box),XtWindow(fw->icon_box),0,0,0,0,True);
     break;
   }
 
@@ -1422,20 +1465,72 @@ void markForUpdate(String path)
       fw->update = True;
 }
 
+#ifdef ENHANCE_PERMS
+/* change of user permission or removing files which
+ * are pointed to by symlinks in our directory are
+ * only detected the hard way :-(, so we must stat
+ * everything to be sure.
+ */
+static Boolean files_changed(FileWindowRec *fw)
+{
+  FileList	fl=fw->files;
+  int		i,statfailed;
+  struct stat	cur;
+  
+  if (chdir(fw->directory)) return True;
+  for (i=0; i<fw->n_files; i++) {
+    statfailed=stat(fl[i]->name,&cur);
+    if (fl[i]->sym_link) {
+	if ( (statfailed && !S_ISLNK(fl[i]->stats.st_mode)) ||
+	     (!statfailed && S_ISLNK(fl[i]->stats.st_mode)) ) {
+	  return True;
+	}
+	if (!statfailed && (cur.st_ctime > fl[i]->stats.st_ctime)) {
+	  return True;
+	}
+    } else {
+    	if ( statfailed || (cur.st_ctime > fl[i]->stats.st_ctime) ) {
+	  return True;
+	}
+    }
+  }
+  return False;
+}
+#endif
+
 void intUpdate()
 {
   FileWindowRec *fw;
   struct stat cur;
+#ifdef ENHANCE_PERMS
+  static int ticks=0;
+
+  ticks++;
+#endif
 
   for (fw = file_windows; fw; fw = fw->next) {
     if (fw->update ||
 	stat(fw->directory, &cur) ||
+#ifdef ENHANCE_PERMS
+	cur.st_ctime > fw->stats.st_ctime || 
+	(ticks==resources.hard_update_ticks && files_changed(fw)))
+#else
 	cur.st_ctime > fw->stats.st_ctime)
+#endif
+#ifdef ENHANCE_SCROLL
+      updateFileDisplay(fw,True);
+#else
       updateFileDisplay(fw);
+#endif
+	
   }
 
   for (fw = file_windows; fw; fw = fw->next)
     fw->update = False;
+
+#ifdef ENHANCE_PERMS
+  if (ticks>=resources.hard_update_ticks) ticks=0;
+#endif
 }
 
 /*-----------------------------------------------------------------------------
@@ -1454,12 +1549,18 @@ void updateStatus(FileWindowRec *fw)
     fillIn(fw->file_items[4]);
     fillIn(fw->file_items[6]);
     fillIn(fw->file_items[10]);
+#ifdef ENHANCE_SELECTION
+    fillIn(fw->file_items[11]);
+#endif
   }else {
     grayOut(fw->file_items[2]);
     grayOut(fw->file_items[3]);
     grayOut(fw->file_items[4]);
     grayOut(fw->file_items[6]);
     grayOut(fw->file_items[10]);
+#ifdef ENHANCE_SELECTION
+    grayOut(fw->file_items[11]);
+#endif
   }
 
   if (fw->display_type == Tree) {         /* incremented the view_item */
@@ -1551,3 +1652,46 @@ void updateStatus(FileWindowRec *fw)
 
   XtVaSetValues(fw->status, XtNlabel, (XtArgVal) s, NULL);
 }
+
+/*----------------------------------------------------------------------------
+  Update a the tick marks of a text options menu and the resources
+  of a text file List to the settings of a FileWindowRec
+-----------------------------------------------------------------------------*/
+
+void updateTxtOpts(FileWindowRec *fw)
+{
+ int i;
+ WidgetList menu_items;
+ static Arg args[]={
+  { XtNshowInode,	(XtArgVal) 0},
+  { XtNshowType,	(XtArgVal) 0},
+  { XtNshowPermissions,	(XtArgVal) 0},
+  { XtNshowLinks,	(XtArgVal) 0},
+  { XtNshowOwner,	(XtArgVal) 0},
+  { XtNshowGroup,	(XtArgVal) 0},
+  { XtNshowLength,	(XtArgVal) 0},
+  { XtNshowDate,	(XtArgVal) 0},
+ };
+
+ if (!XtIsSubclass(fw->icon_box,textFileListWidgetClass)) {
+   XtAppWarning(app_context,"updateTxtOpts(): not a textFileList widget");
+   return;
+ }
+ /* retrieve the menu items */
+ menu_items=fw->option_items;
+
+ i=0;
+ args[i].value=(XtArgVal)fw->showInode; i++;
+ args[i].value=(XtArgVal)fw->showType; i++;
+ args[i].value=(XtArgVal)fw->showPermissions; i++;
+ args[i].value=(XtArgVal)fw->showLinks; i++;
+ args[i].value=(XtArgVal)fw->showOwner; i++;
+ args[i].value=(XtArgVal)fw->showGroup; i++;
+ args[i].value=(XtArgVal)fw->showLength; i++;
+ args[i].value=(XtArgVal)fw->showDate; i++;
+
+ for (i=0; i<XtNumber(args); i++) 
+   if ((Boolean)args[i].value) tick(menu_items[i]); else noTick(menu_items[i]);
+
+ XtSetValues(fw->icon_box,args,XtNumber(args));
+}
