@@ -5,10 +5,13 @@
  (c) A.Graef 1994
  (c) R.Vogelgesang 1994 (`Xfm.BourneShells' stuff)
 
+  modified 7-1997 by strauman@sun6hft.ee.tu-berlin.de to add
+  different enhancements (see README-NEW).
+
  main module for file manager    
 ---------------------------------------------------------------------------*/
 
-#define XFMVERSION "1.3.2"
+#include "FmVersion.h"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -17,6 +20,15 @@
 #include <fcntl.h>
 #include <ctype.h>
 
+#ifdef ENHANCE_USERINFO
+#include <sys/stat.h>
+#include <unistd.h>
+#endif
+
+#ifdef ENHANCE_TRANSLATIONS
+#include <string.h>
+#endif
+
 #ifdef _AIX
 #include <sys/resource.h>
 #endif
@@ -31,6 +43,14 @@
 #include "Am.h"
 #include "Fm.h"
 
+#ifdef ENHANCE_SELECTION
+#include "FmSelection.h"
+#endif
+
+#ifdef ENHANCE_LOG
+#include "FmLog.h"
+#endif
+
 #define XtRDisplayType "DisplayType"
 #define XtRSortType "SortType"
 
@@ -144,6 +164,14 @@ static XtResource resource_list[] = {
       XtOffsetOf(Resources, show_perms), XtRImmediate, (XtPointer) True },
   { "showLength", "ShowLength", XtRBoolean, sizeof(Boolean),
       XtOffsetOf(Resources, show_length), XtRImmediate, (XtPointer) True },
+  { "showInode", "ShowInode", XtRBoolean, sizeof(Boolean),
+      XtOffsetOf(Resources, show_inode), XtRImmediate, (XtPointer) True },
+  { "showType", "ShowType", XtRBoolean, sizeof(Boolean),
+      XtOffsetOf(Resources, show_type), XtRImmediate, (XtPointer) True },
+  { "showGroup", "ShowGroup", XtRBoolean, sizeof(Boolean),
+      XtOffsetOf(Resources, show_group), XtRImmediate, (XtPointer) True },
+  { "showLinks", "ShowLinks", XtRBoolean, sizeof(Boolean),
+      XtOffsetOf(Resources, show_links), XtRImmediate, (XtPointer) True },
   { "defaultDisplayType", "DefaultDisplayType", XtRDisplayType, 
       sizeof(DisplayType), XtOffsetOf(Resources, default_display_type),
       XtRImmediate, (XtPointer) Icons },
@@ -159,12 +187,42 @@ static XtResource resource_list[] = {
   { "updateInterval", "UpdateInterval", XtRInt, sizeof(int),
       XtOffsetOf(Resources, update_interval), XtRImmediate,
       (XtPointer) 10000 },
+#ifdef ENHANCE_PERMS
+  { "hardUpdateTicks", "HardUpdateTicks", XtRInt, sizeof(int),
+      XtOffsetOf(Resources, hard_update_ticks), XtRImmediate, (XtPointer)0 },
+#endif
   { "defaultEditor", "DefaultEditor", XtRString, sizeof(String),
       XtOffsetOf(Resources, default_editor), XtRString, NULL },
   { "defaultViewer", "DefaultViewer", XtRString, sizeof(String),
       XtOffsetOf(Resources, default_viewer), XtRString, NULL },
+  { "defaultXterm", "DefaultXterm", XtRString, sizeof(String),
+      XtOffsetOf(Resources, default_xterm), XtRString, NULL },
   { "BourneShells", "ShellList", XtRString, sizeof(String),
       XtOffsetOf(Resources, sh_list), XtRString, NULL },
+#ifdef ENHANCE_USERINFO
+  { "shellCommand", "ShellCommand", XtRString, sizeof(String),
+      XtOffsetOf(Resources, sh_cmd), XtRString, NULL },
+#endif
+#ifdef ENHANCE_HISTORY
+  { "historyMaxN", "HistoryMaxN", XtRInt, sizeof(int),
+      XtOffsetOf(Resources, history_max_n), XtRImmediate, (XtPointer)30 },
+#endif
+#ifdef ENHANCE_TRANSLATIONS
+  { "appDefsVersion","AppDefsVersion",XtRString,sizeof(String),
+      XtOffsetOf(Resources, app_defs_version), XtRString, (XtPointer)"NOT SET"},
+#endif
+#ifdef XPM
+#ifdef ENHANCE_CMAP
+  { "colorCloseness","ColorCloseness",XtRInt,sizeof(int),
+      XtOffsetOf(Resources, color_closeness), XtRImmediate, (XtPointer)40000},
+#endif
+#endif
+#ifdef ENHANCE_SELECTION
+  { "highlightColor","HighlightColor",XtRPixel,sizeof(Pixel),
+      XtOffsetOf(Resources, highlight_pixel), XtRString, (XtPointer)XtDefaultForeground},
+  { "selectionPathsSeparator","SelectionPathsSeparator",XtRString,sizeof(String),
+      XtOffsetOf(Resources, selection_paths_separator), XtRString, (XtPointer)" "},
+#endif
 };
 
 /*---------------------------------------------------------------------------
@@ -172,30 +230,34 @@ static XtResource resource_list[] = {
 ---------------------------------------------------------------------------*/
 
 static String fallback_resources[] = {
-  "*Command.cursor : hand2",
-  "*MenuButton.cursor : hand2",
-  "*viewport.forceBars: true",
-  "*popup form*bitmap.borderWidth : 0",
-  "*popup form*label.borderWidth : 0",
-  "*button box.orientation : horizontal",
-  "*button box.borderWidth: 0",
-  "*file window*viewport.borderWidth: 0",
-  "*viewport.icon box*Label.borderWidth : 0",
-  "*viewport.icon box.Command.borderWidth : 0",
-  "*viewport.icon box.Form.borderWidth : 0",
-  "*viewport.icon box*Toggle.borderWidth : 1",
-  "*chmod*Label.borderWidth : 0",
-  "*info*Label.borderWidth : 0",
-  "*error*Label.borderWidth : 0",
-  "*confirm*Label.borderWidth : 0",
-  "*Text*translations : #override \\n\
+#ifdef ENHANCE_TRANSLATIONS
+  "Xfm.appDefsVersion: \"\"",
+#endif
+  "Xfm*Command.cursor : hand2",
+  "Xfm*MenuButton.cursor : hand2",
+  "Xfm*popup form*bitmap.borderWidth : 0",
+  "Xfm*popup form*label.borderWidth : 0",
+  "Xfm*button box.orientation : horizontal",
+  "Xfm*button box.borderWidth: 0",
+  "Xfm*file window*viewport.borderWidth: 0",
+  "Xfm*viewport.icon box*Label.borderWidth : 0",
+  "Xfm*viewport.icon box.Command.borderWidth : 0",
+  "Xfm*viewport.icon box.Form.borderWidth : 0",
+  "Xfm*viewport.icon box*Toggle.borderWidth : 1",
+  "Xfm*chmod*Label.borderWidth : 0",
+  "Xfm*info*Label.borderWidth : 0",
+  "Xfm*error*Label.borderWidth : 0",
+  "Xfm*confirm*Label.borderWidth : 0",
+#ifndef ENHANCE_TXT_FIELD
+  "Xfm*Text*translations : #override \\n\
     <Key>Return: no-op() \\n\
     <Key>Linefeed : no-op() \\n\
     Ctrl<Key>J : no-op() \\n",
-  "*Text*baseTranslations : #override \\n\
+  "Xfm*Text*baseTranslations : #override \\n\
     <Key>Return: no-op() \\n\
     <Key>Linefeed : no-op() \\n\
     Ctrl<Key>J : no-op() \\n",
+#endif
 NULL,
 };
 
@@ -352,13 +414,75 @@ void init_arg0flag()
   }
 }
 
+#ifdef ENHANCE_USERINFO
+
+/* 
+ * test if the file named '*name' is executable
+ * *name ist copied to ui->shell in this case.
+ *
+ * return values:
+ *   0	  ok (copy *name to ui->shell)
+ *   1    path not a regular file
+ *   2	  regular file, but no permission to execute
+ */
+
+static int shellpath_not_valid(char *name,UserInfo *ui)
+{
+  struct stat	buf;
+  int		rval;
+
+  if (stat(name,&buf) || ! S_ISREG(buf.st_mode)) rval=1;
+  else if (ui->uid==buf.st_uid) rval=(buf.st_mode & S_IXUSR ? 0:2); 
+  else if (ui->gid==buf.st_gid) rval=(buf.st_mode & S_IXGRP ? 0:2); 
+  else rval = (buf.st_mode & S_IXOTH ? 0:2);
+
+  if (rval==2) fprintf(stderr,"%s exists, no permission to execute\n",name);
+  if (rval==0) strcpy(ui->shell,name);
+  return rval;
+}
+
+/* 
+ * retrieve the name of shell form the
+ * application resources and put it into ui->shell.
+ * if this fails, try to read the environment var SHELL
+ * and finally use "/bin/sh"
+ *
+ * returns 0 if no executable shell was found.
+ */
+static int get_sh_from_resources(UserInfo *ui)
+{
+  char 	*p;
+
+  if (0!=(p=resources.sh_cmd)) {
+    if (!shellpath_not_valid(p,ui)) return 1;
+  }
+
+  if (!(p = getenv("SHELL"))) p="/bin/sh";
+
+  return ! shellpath_not_valid(p,ui);
+}
+#endif
+
 /*---------------------------------------------------------------------------
   Main function
 ---------------------------------------------------------------------------*/
 
+#ifdef ENHANCE_TXT_FIELD
+void noWarnings(String msg)
+{
+}
+#endif
+
+#ifdef ENHANCE_TRANSLATIONS
+static char app_ver0[100], *app_ver;
+#endif
+
 void main(int argc, char *argv[])
 {
   char *s;
+#ifdef ENHANCE_TRANSLATIONS
+  char *p;
+#endif
 
   progname = argv[0];
 
@@ -366,6 +490,7 @@ void main(int argc, char *argv[])
   user.uid = getuid();
   user.gid = getgid();
 
+#ifndef ENHANCE_USERINFO /* read app-resources first */
   if ((s = getenv("HOME")))
     strcpy(user.home,s);
   else
@@ -375,6 +500,7 @@ void main(int argc, char *argv[])
     strcpy(user.shell,s);
   else
     strcpy(user.shell,"/bin/sh");
+#endif
 
   user.umask = umask(0);
   umask(user.umask);
@@ -406,6 +532,28 @@ void main(int argc, char *argv[])
     exit(0);
   }
 
+#ifdef ENHANCE_USERINFO
+  getwd(user.cwd);
+  if (!(s = getenv("HOME"))) s=user.cwd;
+  strcpy(user.home,s);
+
+  if (!get_sh_from_resources(&user)) {
+	fprintf(stderr, "%s%s\n",
+			"no executable shell found in ",
+			"(resources, environment, /bin/sh)");
+	exit(1);
+  }
+#endif
+
+  /* initialise the communications module */
+  initComms();
+
+#ifdef ENHANCE_LOG
+  /* we need the 'tick' bitmap already here */
+  readBitmaps();
+  (void)FmCreateLog(aw.shell,resources.button_font);
+#endif
+
   /* default is to launch both application and file manager */
   if (!resources.appmgr && !resources.filemgr)
     resources.appmgr = resources.filemgr = True;
@@ -426,14 +574,13 @@ void main(int argc, char *argv[])
   sigterm.sa_flags = 0;
   sigaction(SIGTERM,&sigterm,NULL);
 
-  /* initialise the communications module */
-  initComms();
-
   /* check the user's shell; needs signal handlers (to avoid a zombie) */
   init_arg0flag();
 
+#ifndef ENHANCE_LOG
   /* create all the bitmaps & cursors needed */
   readBitmaps();
+#endif
   
   /* Set the icon for the application manager */
   XtVaSetValues(aw.shell, XtNiconPixmap, bm[APPMGR_BM], NULL);
@@ -442,26 +589,43 @@ void main(int argc, char *argv[])
   /* create the main popup shells */
   createMainPopups();
 
-  /* initialise the applications module & create the window */
-  if (resources.appmgr) {
-    strcpy(resources.app_file, resources.app_file_r);
-    fnexpand(resources.app_file);
-    strcpy(resources.main_app_file, resources.app_file);
-    strcpy(resources.app_dir, resources.app_dir_r);
-    fnexpand(resources.app_dir);
-    strcpy(resources.app_clip, resources.app_clip_r);
-    fnexpand(resources.app_clip);
-    readApplicationData(resources.app_file);
-    createApplicationWindow();
-    createApplicationDisplay();
-    XtRealizeWidget(aw.shell);
-    XSetWMProtocols(XtDisplay(aw.shell), XtWindow(aw.shell),
-		    &wm_delete_window, 1);
-    XtAddEventHandler(aw.shell, (EventMask)0L, True,
-		      (XtEventHandler)clientMessageHandler, (XtPointer)NULL);
+#ifdef ENHANCE_TRANSLATIONS
+  /* we don't want to define all the translations as fallback_resources
+   * rather we give up if the application defaults are not properly
+   * installed. (appDefsVersion should be set by the app. defs. file to
+   * the version nr. matching XFMVERSION)
+   */
+  /*strip eventual " */
+  if (resources.app_defs_version)
+    strcpy(app_ver0, resources.app_defs_version);
+  else
+    app_ver0[0] = '\0';
+  app_ver = app_ver0;
+  if (*app_ver==(char)'"') app_ver++;
+  if (p=strrchr(app_ver,'"')) *p=0;
+  if (strcmp(app_ver,XFMVERSION)) {
+    if (*app_ver ==(char)0)
+	error("Sorry: Appl. Defaults Not Found",
+	      "check XFILESEARCHPATH and friends");
+    else {
+	p=XtMalloc(128);
+	sprintf(p,"found: %s, needed: %s",
+	        app_ver,XFMVERSION);
+	error("Appl. Defaults: wrong version number",p);
+        XTFREE(p);
+    }
+    exit(1);
   }
+#endif
 
-  /* initialise the file windows module & create a file window */
+  /* construct the config file names */
+  strcpy(resources.app_file, resources.app_file_r);
+  fnexpand(resources.app_file);
+  strcpy(resources.main_app_file, resources.app_file);
+  strcpy(resources.app_dir, resources.app_dir_r);
+  fnexpand(resources.app_dir);
+  strcpy(resources.app_clip, resources.app_clip_r);
+  fnexpand(resources.app_clip);
   strcpy(resources.cfg_file, resources.cfg_file_r);
   fnexpand(resources.cfg_file);
   strcpy(resources.dev_file, resources.dev_file_r);
@@ -470,15 +634,54 @@ void main(int argc, char *argv[])
   strcpy(resources.magic_file, resources.magic_file_r);
   fnexpand(resources.magic_file);
 #endif
+
+  /* check for config dir and run xfm.install if appropriate */
+  if (!exists(resources.app_dir))
+    if (confirm("It appears that you are running xfm for the first time.",
+		"Would you like me to run the xfm.install script now",
+		"to create your personal setup?"))
+      system("xfm.install -n");
+    else if (aborted)
+      exit(1);
+
+  /* initialise the applications module & create the window */
+  if (resources.appmgr) {
+    readApplicationData(resources.app_file);
+    createApplicationWindow();
+    createApplicationDisplay();
+    XtRealizeWidget(aw.shell);
+    setApplicationWindowName();
+    XSetWMProtocols(XtDisplay(aw.shell), XtWindow(aw.shell),
+		    &wm_delete_window, 1);
+    XtAddEventHandler(aw.shell, (EventMask)0L, True,
+		      (XtEventHandler)clientMessageHandler, (XtPointer)NULL);
+  }
+
+#ifdef ENHANCE_SELECTION
+  /* initialize the selection ICCM */
+  FmSelectionInit(XtDisplay(aw.shell));
+#endif
+
+  /* initialise the file windows module & create a file window */
   initFileWindows();
   if (resources.filemgr)
+#ifdef ENHANCE_USERINFO
+    newFileWindow(user.cwd,resources.initial_display_type,False);
+#else
     newFileWindow(user.home,resources.initial_display_type,False);
+#endif
   resources.init_geometry = NULL;
 
   /* start up window refresh timer */
   if (resources.update_interval > 0)
     XtAppAddTimeOut(app_context, resources.update_interval, timeoutCb, NULL);
 
+#ifdef ENHANCE_TXT_FIELD
+  /* those 'text cursor out of range' messages from the TextField widget kept
+     annoying me, so I switch Xt warnings off here --AG */
+  XtAppSetWarningHandler(app_context, noWarnings);
+#endif
+
   /* collect & process events */
   XtAppMainLoop(app_context);
 }
