@@ -7,6 +7,10 @@
   - default values for parameters added in varPopup() by Brian King
     (ender@ee.WPI.EDU), integrated Mar 24 1995, AG
 
+  modified 7-1997 by strauman@sun6hft.ee.tu-berlin.de to add
+  different enhancements (see README-NEW).
+
+
   General utility functions for creating menus, buttons, questions,
   and functions for desensetising and 'ticking' menu entries.
 -----------------------------------------------------------------------------*/
@@ -22,7 +26,12 @@
 #include <X11/Xaw/Form.h>
 #include <X11/Xaw/Label.h>
 #include <X11/Xaw/Box.h>
+#ifdef ENHANCE_TXT_FIELD
+#include "TextField.h"
+#include "FocusForm.h"
+#else
 #include <X11/Xaw/AsciiText.h>
+#endif
 
 #include "Am.h"
 #include "Fm.h"
@@ -85,8 +94,12 @@ static Arg text_args[] = {
   { XtNbottom, XtChainTop },
   { XtNleft, XtChainLeft },
   { XtNright, XtChainRight },
+#ifndef ENHANCE_TXT_FIELD
   { XtNeditType, XawtextEdit },
   { XtNtype, XawAsciiString },
+#else
+  { XtNfocusInterest, True },
+#endif
   { XtNuseStringInPlace, (XtArgVal) True },
 };
 
@@ -264,27 +277,49 @@ Widget *createButtons(ButtonList buttons, Cardinal n_buttons, Widget parent,
 /*            n_questions :  Number of questions                             */
 /*            buttons     :  A set of buttons to put at the bottom           */
 /*            n_buttons   :  Number of buttons                               */
+/*            dflt_button :  Default button to use when the popup is closed  */
+/*                           (WM_DELETE_WINDOW message)                      */
 /*                                                                           */
 /* Create a popup questionaire with a bitmap to the left (or none), several  */
 /* questions (each consisting of a label and a text area) to the right of    */
 /* the bitmap, and a set of buttons underneath all this.                     */
 /*****************************************************************************/
 
+static void clmHandler(Widget w, XtPointer cld, XEvent *ev, Boolean *ctd)
+{
+  XtCallbackProc cb = (XtCallbackProc)cld;
+  XClientMessageEvent *cme = (XClientMessageEvent *)ev;
+  if (ev->type != ClientMessage || cme->message_type != wm_protocols)
+    return;
+  if (cb) cb(w, NULL, NULL);
+}
+
 Widget createPopupQuestions(String name, String title, Pixmap bitmap,
 			    QuestionList questions, Cardinal n_questions,
-			    ButtonList buttons, Cardinal n_buttons)
+			    ButtonList buttons, Cardinal n_buttons,
+			    Cardinal dflt_button)
 {
   int i, l;
   Widget form_widget, box_widget, bitmap_widget = NULL, shell,
     vert = NULL, horiz = NULL;
-
+#ifdef ENHANCE_POP_ACCEL
+  Widget *button_widgets;
+#ifndef ENHANCE_TRANSLATIONS
+  int flag;
+#endif
+#endif
   /* create popup shell */
   shell_args[0].value = (XtArgVal) title;
   shell = XtCreatePopupShell(name, transientShellWidgetClass, aw.shell,
 			     shell_args, XtNumber(shell_args) );
 
   /* create form */
+  form_args[0].value = PADDING;
+#ifdef ENHANCE_TXT_FIELD
+  form_widget = XtCreateManagedWidget("popup form", focusFormWidgetClass, shell, 
+#else
   form_widget = XtCreateManagedWidget("popup form", formWidgetClass, shell, 
+#endif
 				      form_args, XtNumber(form_args) );
 
   /* create bitmap */
@@ -294,6 +329,8 @@ Widget createPopupQuestions(String name, String title, Pixmap bitmap,
 	  form_widget, bitmap_args, XtNumber(bitmap_args));
   }
 
+  label_args[5].value = XtJustifyRight;
+
   /* Find width of label */
   label_args[3].value = (XtArgVal) 0;
   for (i=0; i<n_questions; i++) {
@@ -320,7 +357,11 @@ Widget createPopupQuestions(String name, String title, Pixmap bitmap,
     text_args[2].value = (XtArgVal) questions[i].value;
     text_args[3].value = (XtArgVal) questions[i].length;
     vert = questions[i].widget = XtCreateManagedWidget("text", 
+#ifdef ENHANCE_TXT_FIELD
+	   textFieldWidgetClass, form_widget, text_args, XtNumber(text_args));
+#else
 	   asciiTextWidgetClass, form_widget, text_args, XtNumber(text_args));
+#endif
   }
 
   if (buttons != NULL) {
@@ -328,10 +369,49 @@ Widget createPopupQuestions(String name, String title, Pixmap bitmap,
     button_box_args[1].value = (XtArgVal) vert;
     box_widget = XtCreateManagedWidget("button box", boxWidgetClass, 
 	        form_widget, button_box_args, XtNumber(button_box_args));
+#ifdef ENHANCE_POP_ACCEL
+    button_widgets =
+#endif
     createButtons(buttons, n_buttons, box_widget, NULL);
+
+#ifdef ENHANCE_POP_ACCEL
+    for (i = 0; i<n_questions; i++) {
+    	XtInstallAllAccelerators(questions[i].widget,form_widget);
+    }
+    XtInstallAllAccelerators(form_widget,form_widget);
+#ifndef ENHANCE_TRANSLATIONS
+    for (i=0;i<n_buttons;i++) {
+	/* This allows the user to trigger the buttons
+	 * by hitting keys anywhere in the form
+	 * provided translations and accelerators are set
+	 * correctly in the resource file
+	 */
+      flag=0;
+      if (!strcmp(buttons[i].name,"ok")) flag=1; 
+      else if (!strcmp(buttons[i].name,"cancel")) flag=2;
+      if (flag) {
+	XtVaSetValues(button_widgets[i],
+	  XtNaccelerators,
+	  (flag==1?
+	    "<Key>Return:set()notify()unset()\n <Key>Linefeed:set()notify()unset()\n":
+	    "<Key>Escape:set()notify()unset()\n"),
+	  NULL);
+      }
+    }
+#endif
+#endif
   }
 
   XtRealizeWidget(shell);
+  XSetWMProtocols(XtDisplay(shell), XtWindow(shell),
+		  &wm_delete_window, 1);
+  if (buttons && 0 <= dflt_button && dflt_button < n_buttons)
+    XtAddEventHandler(shell, (EventMask)0L, True,
+		      (XtEventHandler)clmHandler,
+		      (XtPointer)buttons[dflt_button].callback);
+  else
+    XtAddEventHandler(shell, (EventMask)0L, True,
+		      (XtEventHandler)clmHandler, (XtPointer)NULL);
 
   return shell;
 }
@@ -541,7 +621,8 @@ char *varPopup(Pixmap icon_bm, char *action)
 
     dialog = createPopupQuestions("dialog", "Parameter Dialog", icon_bm,
 				  dialog_questions, n_vars, dialog_buttons,
-				  XtNumber(dialog_buttons));
+				  XtNumber(dialog_buttons),
+				  XtNumber(dialog_buttons)-1);
     popupByCursor(dialog, XtGrabExclusive);
 
     dialog_flag = DontKnow;
@@ -576,3 +657,149 @@ char *varPopup(Pixmap icon_bm, char *action)
   }
 }
 
+/*---------------------------------------------------------------------------*/
+
+/* variant of createPopupQuestions which shows some lines of left-justified
+   text below a pixmap */
+
+static Widget createPopupText(String name, String title, Pixmap bitmap,
+			      String *lines, Cardinal n_lines,
+			      ButtonList buttons, Cardinal n_buttons,
+			      Cardinal dflt_button)
+{
+  int i, l;
+  Widget form_widget, box_widget, bitmap_widget = NULL, shell,
+    vert = NULL;
+#ifdef ENHANCE_POP_ACCEL
+  Widget *button_widgets;
+#ifndef ENHANCE_TRANSLATIONS
+  int flag;
+#endif
+#endif
+  /* create popup shell */
+  shell_args[0].value = (XtArgVal) title;
+  shell = XtCreatePopupShell(name, transientShellWidgetClass, aw.shell,
+			     shell_args, XtNumber(shell_args) );
+
+  /* create form */
+  form_args[0].value = 5;
+#ifdef ENHANCE_TXT_FIELD
+  form_widget = XtCreateManagedWidget("popup form", focusFormWidgetClass, shell, 
+#else
+  form_widget = XtCreateManagedWidget("popup form", formWidgetClass, shell, 
+#endif
+				      form_args, XtNumber(form_args) );
+  /* create bitmap */
+  if (bitmap != None) {
+    bitmap_args[2].value = (XtArgVal) bitmap;
+    vert = bitmap_widget = XtCreateManagedWidget("bitmap", labelWidgetClass,
+	  form_widget, bitmap_args, XtNumber(bitmap_args));
+  }
+
+  label_args[5].value = XtJustifyLeft;
+
+  /* Find width of label */
+  label_args[3].value = (XtArgVal) 0;
+  for (i=0; i<n_lines; i++) {
+    l = XTextWidth(resources.label_font, lines[i], 
+		   strlen(lines[i]));
+    if (l > label_args[3].value)
+      label_args[3].value = l;
+  }
+
+  for (i = 0; i<n_lines; i++) {
+    label_args[0].value = (XtArgVal) NULL;
+    label_args[1].value = (XtArgVal) vert;
+    label_args[2].value = (XtArgVal) lines[i];
+    vert = XtCreateManagedWidget("label", labelWidgetClass, form_widget,
+				 label_args, XtNumber(label_args));
+  }
+
+  if (buttons != NULL) {
+    button_box_args[0].value = (XtArgVal) NULL;
+    button_box_args[1].value = (XtArgVal) vert;
+    box_widget = XtCreateManagedWidget("button box", boxWidgetClass, 
+	        form_widget, button_box_args, XtNumber(button_box_args));
+#ifdef ENHANCE_POP_ACCEL
+    button_widgets =
+#endif
+    createButtons(buttons, n_buttons, box_widget, NULL);
+
+#ifdef ENHANCE_POP_ACCEL
+    XtInstallAllAccelerators(form_widget,form_widget);
+#ifndef ENHANCE_TRANSLATIONS
+    for (i=0;i<n_buttons;i++) {
+	/* This allows the user to trigger the buttons
+	 * by hitting keys anywhere in the form
+	 * provided translations and accelerators are set
+	 * correctly in the resource file
+	 */
+      flag=0;
+      if (!strcmp(buttons[i].name,"ok")) flag=1; 
+      else if (!strcmp(buttons[i].name,"cancel")) flag=2;
+      if (flag) {
+	XtVaSetValues(button_widgets[i],
+	  XtNaccelerators,
+	  (flag==1?
+	    "<Key>Return:set()notify()unset()\n <Key>Linefeed:set()notify()unset()\n":
+	    "<Key>Escape:set()notify()unset()\n"),
+	  NULL);
+      }
+    }
+#endif
+#endif
+  }
+
+  XtRealizeWidget(shell);
+  XSetWMProtocols(XtDisplay(shell), XtWindow(shell),
+		  &wm_delete_window, 1);
+  if (buttons && 0 <= dflt_button && dflt_button < n_buttons)
+    XtAddEventHandler(shell, (EventMask)0L, True,
+		      (XtEventHandler)clmHandler,
+		      (XtPointer)buttons[dflt_button].callback);
+  else
+    XtAddEventHandler(shell, (EventMask)0L, True,
+		      (XtEventHandler)clmHandler, (XtPointer)NULL);
+
+  return shell;
+}
+
+/*---------------------------------------------------------------------------*/
+
+static Widget about = None;
+static Pixmap logo_bm;
+
+static void aboutOkCb(Widget w, XtPointer client_data, 
+		      XtPointer call_data)
+{
+  XtPopdown(about);
+}
+
+/*---------------------------------------------------------------------------*/
+
+static ButtonRec about_buttons[] = {
+  { "ok", "Ok", (FmCallbackProc *) aboutOkCb },
+};
+
+static char ver_str[100];
+static String about_text[] = {
+  ver_str,
+  "Written by Simon Marlow, Albert Graef, Till Straumann, et al",
+  "Copying policy: GPL Version 2 or later",
+};
+
+#include "FmVersion.h"
+
+void aboutPopup(void)
+{
+  if (about == None) {
+    sprintf(ver_str, "This is xfm version %s", XFMVERSION);
+    logo_bm = readIcon("XfmLogo.xpm", 0, 0);
+    about = createPopupText("about", "About xfm", logo_bm,
+			    about_text, XtNumber(about_text), about_buttons,
+			    XtNumber(about_buttons),
+			    XtNumber(about_buttons)-1);
+  }
+  if (!XtIsManaged(about))
+    popupByCursor(about, XtGrabNone);
+}
